#
# Copyright (C) Slava Kozlov 2002 - All rights reserved
#
# TWiki extension to TablePlugin module: Coloring.pm
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details, published at 
# http://www.gnu.org/copyleft/gpl.html
#


package TWiki::Plugins::TablePlugin::Coloring;

use strict;
use TWiki::Func;
use TWiki::Plugins::TablePlugin qw(@dataBg @bgFilter @cellColoring);

use vars qw(%coloringSubMap $BLACK $WHITE $MAXCOL @EXPORT
	&cellcolor &registerColoring);

# export core function, comment out if you like looser exports
use base qw(Exporter);
@EXPORT = qw(cellcolor registerColoring);


$WHITE = "#ffffff";
$BLACK = "#000000";
$MAXCOL = 0xFF; # maximum value for a single color component

# cellcoloring value to routine map

%coloringSubMap = (
	'horizontal'	=> \&horizontalColoring,
	'vertical' 	=> \&verticalColoring,	
	'diagonal' 	=> \&diagonalColoring,
	'chessboard' 	=> \&chessboardColoring,	
	'highlight' 	=> \&highlightColoring,	
	'filter' 	=> \&filterColoring,
	'filterbycol' 	=> \&filterbycolColoring,	
	'filterbyrow' 	=> \&filterbyrowColoring,	
	'filtermix' 	=> \&filtermixColoring,
	'filtermultimix' => \&filtermultmixColoring,	
	'gradient' 	=> \&gradientColoring,	
	'gradientcol' 	=> \&gradientcolColoring,	
	'gradientdia' 	=> \&gradientdiaColoring,		
	'gradientcenter' => \&gradientcenterColoring,		
	'border' 	=> \&borderColoring	
);

# management function
# can add more coloring functions (need TWiki interface)

sub registerColoring
{
	my ($name, $function) = @_;
	$coloringSubMap{$name} = $function;
}

# one day I'll figure out perl's exporting rules (sigh)
# maybe this is actually better than direct package access

# some data mapping functions

sub dataBg {
	return \@TWiki::Plugins::TablePlugin::dataBg;
}
sub bgFilter {
	return \@TWiki::Plugins::TablePlugin::bgFilter;
}
sub cellColoring {
	return \@TWiki::Plugins::TablePlugin::cellColoring;
}


# function

sub cellcolor {
	#TWiki::Func::writeDebug("cellcolor: ".$#{dataBg()});
	
	return $WHITE unless (@{dataBg()}); # is this redundant?    
#	TWiki::Func::writeDebug("cellcoloring: ".${cellColoring()}[0]);
#	TWiki::Func::writeDebug("cellcoloring func: ".$coloringSubMap{${cellColoring()}[0]} );
	
	my $coloringSub = $coloringSubMap{${cellColoring()}[0]};

#	my $r = &$coloringSub(@_);
#	TWiki::Func::writeDebug("cellcolor: ".$r);
#	return $r;
	
    	return &$coloringSub(@_);
}

######################################################
######  Colorings which alternate the colors in dataBg

### alternate rows

sub horizontalColoring {
	my ($cell, $row, $rowCount) = @_;
	return _modDataColor($rowCount);
}

### alternate cols

sub verticalColoring {
	my ($cell, $row, $rowCount, $colCount) = @_;
	return _modDataColor($colCount);
}

### alternate on the down-right diagonal
### why would anyone want this?

sub diagonalColoring {
	my ($cell, $row, $rowCount, $colCount) = @_;
	return _modDataColor($colCount + $rowCount);
}

### chessboard pattern - can do with two colors and diagonal

sub chessboardColoring {
	my ($cell, $row, $rowCount, $colCount) = @_;
	return _modDataColor( ($colCount + $rowCount) % 2);
}

#--------- helper function

### mod helper function
sub _modDataColor {
	return ${dataBg()}[$_[0] % ($#{dataBg()} + 1)];
}


######################################################
######  Colorings that color certain rows/columns

### color the border with two colors (remember header is considered a row!)

sub borderColoring {
	my ($cell, $row, $rowCount, $colCount, $curTable) = @_;
	return _modDataColor(
		$rowCount == 0 ||
		$colCount == 0 ||
		$rowCount == $#{$curTable} ||
		$colCount == $#{$row}
	);
}

### Color a row, a column or intersection
#
# this is a little complicated interface,
# depending on two things:
#
# A) how many entries in dataBg:
#
# 1) 0 colors: colors in white
# 2) 1 color : colors both row and col -> 1st, inter mixed with black, default white
# 3) 2 colors: row -> 1st, col -> 1st, intersection mixed, default 2nd
# 4) 3 colors: row -> 1st, col -> 2nd, intersection mixed, default 3rd
# 5) 4 colors: row -> 1st, col -> 2nd, intersection 3rd, default 4th
#
# B) highlight's attributes
# 
# 1) given both row and col (eg, highlight:4:3) colors both, mixes intersection
# 2) given only row (highlight:4) highlights row
# 3) given only col (highlight::4) highlights col
#
# NB: always defaults to last entry in dataBg
#
# NB: haven't tested/implemented all these out (is A.5 done?)

sub highlightColoring {
	my ($cell, $row, $rowCount, $colCount, $curTable) = @_;
	return $WHITE unless $#{dataBg()} > -1;
	my $interRow = ${cellColoring()}[1] || 0; $interRow--;
	my $interCol = ${cellColoring()}[2] || 0; $interCol--;
	($rowCount == $interRow) && ($colCount == $interCol)
		&& return  _multiMixColors(${dataBg()}[0], ${dataBg()}[1] || $BLACK);
	($colCount == $interCol) && return ${dataBg()}[0];
	($rowCount == $interRow) && return (${dataBg()}[1] || ${dataBg()}[0]);
	return (${dataBg()}[$#{dataBg()}] || $WHITE );
}


### colors a cell by whether it matches a filter value

sub filterColoring {
	my ($cell) = @_;
	return _colorFilterText($cell);
}	

### colors a cell by whether a certain row matches a filter value
# eg, cellcoloring:3 (3rd row matching)
# tested?
sub filterbyrowColoring {
	my ($cell, $row, $rowCount, $colCount, $curTable) = @_;

        # get cellColoring attribute qualifier
        my $filterRow = ${cellColoring()}[1] || 1;

        # grab the cell from the filterRow ( -1 because humans don't start counting at zero)
        my $filterCell = $curTable->[$filterRow - 1]->[$colCount]->[0];
	return _colorFilterText($filterCell);
}

### colors a cell by whether a certain col matches a filter value
# eg, cellcoloring:3 (3rd col matching)

sub filterbycolColoring {
	my ($cell, $row) = @_;

        # get cellColoring attribute qualifier
        my $filterColumn = ${cellColoring()}[1] || 1;

        # grab the cell from the filterColumn ( -1 because humans don't start counting at zero)
        my $filterCell = $row->[$filterColumn - 1]->[0];
	return _colorFilterText($filterCell);
}

#----------- helper functions for filtering

# loops thru filters and applies given functor to matches
#
sub filterLoop {
  my ( $closure, $text ) = @_;
  for (my $color = 0; $color < $#{bgFilter()} + 1 ; $color++) {
    &$closure($color) if ($text =~ /${bgFilter()}[$color]/ );
  }
}

# given string looks thru filters to find match
# returns dataBg's matching color
#
# NB: finds first match not last (ie, defaults should be at back of dataBg)
#
sub _colorFilterText {
	my ($cellText) = @_;
	return 0 unless (@{bgFilter()});
	my $color= -1;
	my $found = 0;
	my $sub = sub {
		$color = shift unless $found; # only get first match
		$found = 1;
	};
  	filterLoop( $sub, $cellText);
	$color = $#{dataBg()} unless ($found) ; # ($color > $#{dataBg()} );
	return ${dataBg()}[$color];
}

###################################
### Color mixing colorings

#### gradient coloring horizontally (only between first two colors in dataBg)

sub gradientColoring {
	my ($cell, $row, $rowCount, $colCount, $curTable) = @_;
	return _gradColor($rowCount, $#{$curTable});
}

#### gradient coloring vertically (only between first two colors in dataBg)

sub gradientcolColoring {
	my ($cell, $row, $rowCount, $colCount, $curTable) = @_;
	return _gradColor($colCount, $#{$row} );
}

#### gradient coloring on the right-down diagonal (only between first two colors in dataBg)

sub gradientdiaColoring {
	my ($cell, $row, $rowCount, $colCount, $curTable) = @_;
	return _gradColor($colCount + $rowCount, $#{$curTable} + $#{$row});
}

#### gradient coloring toward the center

sub gradientcenterColoring {
	my ($cell, $row, $rowCount, $colCount, $curTable) = @_;
	return _gradColor(
	   ( abs($colCount - $#{$row}/2) + abs($rowCount - $#{$curTable}/2) ),
	   $#{$row}/2 + $#{$curTable}/2 );
}


# color cell by its position in the interval length ($total)
# and this cell's position in it ($num)
	
sub _gradColor {
	my ($num, $total) = @_;	
	return $WHITE if ($#{dataBg()} == -1);
	return ${dataBg()}[0] if ($#{dataBg()} == 0);
	
	my $A = _colorToTriplet(${dataBg()}[0]);
	my $B = _colorToTriplet(${dataBg()}[1]);

	# mix the right postion between A and B
	return _tripletToColor(
		_doTriplet(sub { ($num) * ($_[1] - $_[0]) / ($total+1) + $_[0] },
		$A,
		$B ));
}

######## Mixing and filtering together

# mixes matched colors (to filterBg) by averaging

sub filtermixColoring {
	my ($cell, $row, $rowCount, $colCount) = @_;
	return _colorMixFilter($cell);
}
# mixes matched colors (to bgFilter) by multiplying

sub filtermultmixColoring {
	my ($cell, $row, $rowCount, $colCount) = @_;
	return _colorMultMixFilterText($cell);
}


# no sure what this done, it "multiplies" all the colors together

sub _colorMultMixFilterText {
	my ($cellText) = @_;
	return 0 unless (@{bgFilter()});
	my @w = (1,1,1); my $finalC = \@w; # mult identity
	my $colorsTotal = 0;

  	# filter routine
  	my $sub = sub {
		my ($color) = @_; 
		# mix in data color by multiplying normalized colors
		my $T = _doTriplet( sub { $_[0] * $_[1] / $MAXCOL},
			  _colorToTriplet(${dataBg()}[$color]), 	
		  	  $finalC );
		$finalC = $T;
  	  	$colorsTotal++;		# record additional match
  	};

  	filterLoop( $sub, $cellText);
  	
	return ${dataBg()}[$#{dataBg()}] unless $colorsTotal;
	return _tripletToColor(  	# multiply 0xFF back in, format & done
	  _doTriplet( sub { $_[0] * $MAXCOL }, $finalC ));
}



# given text return color which
# is average of the colors that match a cell

sub _colorMixFilter {
	my ($cellText) = @_;
	return 0 unless (@{bgFilter()});
	my @zero = (0,0,0); my $finalC = \@zero; # addition identity
	my $colorsTotal = 0;
	
  	# filter routine
  	my $sub = sub {
		my ($color) = @_; 
		my $T = _doTriplet( sub { $_[0] + $_[1] },    # mix in data color
			  _colorToTriplet(${dataBg()}[$color]), 	
		  	  $finalC );
		$finalC = $T;
  	  	$colorsTotal++;		# record additional match
  	};

  	filterLoop( $sub, $cellText);
#TWiki::Func::writeDebug("matches: $cellText - ".$colorsTotal);

	return ${dataBg()}[$#{dataBg()}] unless $colorsTotal; # no matches
	return _tripletToColor(
		_doTriplet( sub { $_[0]/ $colorsTotal ; }, $finalC) ); # normalize triplet, convert to color, & done
}



#######################################

## helper functions for mixing/dealing with colors
# should these be in a seperate package?

# given two colors mixes them by averaging
sub _mixColors{
	my ($a, $b) = @_;
	return _tripletToColor(
		_mixTriplets(_colorToTriplet($a), _colorToTriplet($b)));
}


# given two triplets mixes them by averaging
sub _mixTriplets{
	my ($A, $B) = @_;
	return _doTriplet(sub { ($_[0] + $_[1])/2 }, $A, $B );
}

# given two colors mixes them by norm-multiplying
sub _multiMixColors{
	my ($a, $b) = @_;
	return _tripletToColor(
		_multiMixTriplets(_colorToTriplet($a), _colorToTriplet($b)));
}

# given two triplets, mixes them by norm-multiplication

sub _multiMixTriplets{
	my ($A, $B) = @_;
	return _doTriplet(sub { ($_[0] * $_[1])/$MAXCOL  }, $A, $B);
}


# applies the function /do/ to the triplet refs vectors, A & B
# returns a pointer to a triplet

sub _doTriplet {
	my ($do, $A, $B) = @_;
	my $C;
	foreach my $i (0..2) {
		$C->[$i] = &$do( $A->[$i], (ref $B) ? $B->[$i] : 0 );
	}
	return $C;
}


# converts an HTML color to a RGB triplet pointer

sub _colorToTriplet {
	my ($color) = @_;
	my $n = _colorToNum($color);
	my @t;
	my $a = $n % 0x100;
	$n = ($n - $a) / 0x100;
	my $b = $n % 0x100;
	$n = ($n - $b) / 0x100;
	
	push @t, $n, $b, $a;
	return \@t;
}


# converts an HTML color to a number
# only works with hexadecimal colornames (eg, #ffbb1A")

sub _colorToNum {
	my ($color) = @_;
	$color =~ s/\#/0x/; # make color into Perl hex string
	return eval($color);
}


# converts a ref to a triplet (vector) to a HTML color
sub _tripletToColor {
	my ($V) = @_;
	return "" if ($#{$V} < 2);
	return _numToColor( int($V->[0]+.5) * 0x10000 + int($V->[1]+.5) * 0x100 + $V->[2] );
}

# converts a decimal number to hexidecimal string with pound sign (#)
sub _numToColor {
	return sprintf("#%lx",  $_[0]);
}



1;

